#include <common.h>
#include <asm/io.h>
#include <asm/gpio.h>
#include "bspchip.h"

#define GPIO_RANGE 102

#define RTS_GETFIELD(val, width, gpio)	\
			(((val) >> (gpio)) & ((1 << (width)) - 1))
#define RTS_SETFIELD(reg, field, width, gpio)	(((reg) & \
			(~(((1 << (width)) - 1) <<	(gpio)))) \
			| (((field) & ((1 << (width)) - 1)) << (gpio)))

struct sharepin_cfg_addr {
	int pinl;
	int pinh;
	int pint;
	int pinaddr;
};

struct pinregs {
	int gpio_oe;
	int gpio_value;
	int pullctrl;
	int gpio_int_en;
	int gpio_int;
	int drv_sel;
	int sr_ctrl;
	int pad_cfg;
};

enum {
	GPIO_TYPE_GENERIC,
	GPIO_TYPE_UART0,
	GPIO_TYPE_UART1,
	GPIO_TYPE_UART2,
	GPIO_TYPE_PWM,
	GPIO_TYPE_I2C,
	GPIO_TYPE_SDIO0,
	GPIO_TYPE_SDIO1,
	GPIO_TYPE_SSOR,
	GPIO_TYPE_DMIC,
	GPIO_TYPE_ADDA,
	GPIO_TYPE_I2S,
	GPIO_TYPE_SARADC,
	GPIO_TYPE_MIPITX,
	GPIO_TYPE_LVDS,
	GPIO_TYPE_USBD,
	GPIO_TYPE_USBH,
	GPIO_TYPE_USB3,
};

static struct sharepin_cfg_addr pincfgaddr[] = {
	{.pinl = 0, .pinh = 15,
	 .pint = GPIO_TYPE_GENERIC, .pinaddr = GPIO_OE},
	{.pinl = 16, .pinh = 19,
	 .pint = GPIO_TYPE_UART0, .pinaddr = UART0_GPIO_OE},
	{.pinl = 20, .pinh = 21,
	 .pint = GPIO_TYPE_UART1, .pinaddr = UART1_GPIO_OE},
	{.pinl = 22, .pinh = 23,
	 .pint = GPIO_TYPE_UART2, .pinaddr = UART2_GPIO_OE},
	{.pinl = 24, .pinh = 27,
	 .pint = GPIO_TYPE_PWM, .pinaddr = PWM_GPIO_OE},
	{.pinl = 28, .pinh = 29,
	 .pint = GPIO_TYPE_I2C, .pinaddr = XB2_I2C_GPIO_OE},
	{.pinl = 30, .pinh = 37,
	 .pint = GPIO_TYPE_SDIO0, .pinaddr = SD0_GPIO_OE},
	{.pinl = 38, .pinh = 45,
	 .pint = GPIO_TYPE_SDIO1, .pinaddr = SD1_GPIO_OE},
	{.pinl = 46, .pinh = 58,
	 .pint = GPIO_TYPE_SSOR, .pinaddr = VIDEO_GPIO_OE},
	{.pinl = 59, .pinh = 62,
	 .pint = GPIO_TYPE_DMIC, .pinaddr = DMIC_GPIO_OE},
	{.pinl = 63, .pinh = 66,
	 .pint = GPIO_TYPE_ADDA, .pinaddr = AUDIO_ADDA_GPIO_OE},
	{.pinl = 67, .pinh = 71,
	 .pint = GPIO_TYPE_I2S, .pinaddr = I2S_GPIO_OE},
	{.pinl = 72, .pinh = 75,
	 .pint = GPIO_TYPE_SARADC, .pinaddr = SARADC_GPIO_OE},
	{.pinl = 76, .pinh = 85,
	 .pint = GPIO_TYPE_MIPITX, .pinaddr = MIPI_TX_GPIO_OE},
	{.pinl = 86, .pinh = 95,
	 .pint = GPIO_TYPE_LVDS, .pinaddr = LVDS_GPIO_OE},
	{.pinl = 96, .pinh = 97,
	 .pint = GPIO_TYPE_USBD, .pinaddr = USB0_GPIO_OE},
	{.pinl = 98, .pinh = 99,
	 .pint = GPIO_TYPE_USBH, .pinaddr = USB1_GPIO_OE},
	{.pinl = 100, .pinh = 102,
	 .pint = GPIO_TYPE_USB3, .pinaddr = USB2_GPIO_OE},
};

static struct sharepin_cfg_addr *rts_get_pinaddr(unsigned pin)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(pincfgaddr); i++)
		if (pincfgaddr[i].pinl <= pin && pincfgaddr[i].pinh >= pin)
			return &pincfgaddr[i];
	return 0;
}

static unsigned int rts_gpio_get_field(void __iomem *reg,
				       unsigned int width, unsigned int gpio)
{
	unsigned int val = readl(reg);

	return RTS_GETFIELD(val, width, gpio);
}

static void rts_gpio_set_field(void __iomem *reg,
			       unsigned int field, unsigned int width,
			       unsigned int gpio)
{
	unsigned int val = readl(reg);

	val = RTS_SETFIELD(val, field, width, gpio);
	writel(val, reg);
}

static int rts_gpio_enable(unsigned gpio)
{
	struct pinregs *regs;
	struct sharepin_cfg_addr *sc;
	int bf;

	sc = rts_get_pinaddr(gpio);
	regs = (struct pinregs *)sc->pinaddr;
	bf = gpio - sc->pinl;

	switch (sc->pint) {
	case GPIO_TYPE_GENERIC:
		break;
	case GPIO_TYPE_UART0:
	case GPIO_TYPE_UART1:
	case GPIO_TYPE_UART2:
	case GPIO_TYPE_PWM:
	case GPIO_TYPE_I2C:
	case GPIO_TYPE_I2S:
	case GPIO_TYPE_USB3:
		rts_gpio_set_field(&regs->pad_cfg, 1, 4, bf << 2);
		break;
	case GPIO_TYPE_SDIO0:
	case GPIO_TYPE_SDIO1:
		rts_gpio_set_field(&regs->pad_cfg, 1, 2, 0);
		break;
	case GPIO_TYPE_SARADC:
	case GPIO_TYPE_USBD:
	case GPIO_TYPE_USBH:
		rts_gpio_set_field(&regs->pad_cfg, 2, 4, bf << 2);
		break;
	case GPIO_TYPE_SSOR:
		if (bf >= 11)
			bf = 11;
		bf /= 2;
		if (bf >= 3)
			bf++;
		rts_gpio_set_field(&regs->pad_cfg, 1, 4, bf << 2);
		break;
	case GPIO_TYPE_DMIC:
		bf /= 2;
		rts_gpio_set_field(&regs->pad_cfg, 1, 4, bf << 2);
		break;
	case GPIO_TYPE_MIPITX:
	case GPIO_TYPE_ADDA:
	case GPIO_TYPE_LVDS:
		bf /= 2;
		rts_gpio_set_field(&regs->pad_cfg, 2, 4, bf << 2);
		break;
	default:
		break;
	}
	return 0;
}

int gpio_request(unsigned gpio, const char *label)
{
	if (gpio > GPIO_RANGE)
		return -EINVAL;
	rts_gpio_enable(gpio);
	return 0;
}

int gpio_free(unsigned gpio)
{
	if (gpio > GPIO_RANGE)
		return -EINVAL;
	return 0;
}

int gpio_get_value(unsigned gpio)
{
	struct pinregs *regs;
	struct sharepin_cfg_addr *sc;
	int bf;

	if (gpio > GPIO_RANGE)
		return -EINVAL;
	sc = rts_get_pinaddr(gpio);
	regs = (struct pinregs *)sc->pinaddr;
	bf = gpio - sc->pinl;

	return rts_gpio_get_field(&regs->gpio_value, 1, bf);
}

int gpio_set_value(unsigned gpio, int value)
{
	struct pinregs *regs;
	struct sharepin_cfg_addr *sc;
	int bf;

	if (gpio > GPIO_RANGE)
		return -EINVAL;
	sc = rts_get_pinaddr(gpio);
	regs = (struct pinregs *)sc->pinaddr;
	bf = gpio - sc->pinl;

	if (value)
		set_bit(bf, &regs->gpio_value);
	else
		clear_bit(bf, &regs->gpio_value);
	return 0;
}

int gpio_direction_input(unsigned gpio)
{
	struct pinregs *regs;
	struct sharepin_cfg_addr *sc;
	int bf;

	if (gpio > GPIO_RANGE)
		return -EINVAL;
	sc = rts_get_pinaddr(gpio);
	regs = (struct pinregs *)sc->pinaddr;
	bf = gpio - sc->pinl;

	rts_gpio_set_field(&regs->gpio_oe, 0, 1, bf);

	return 0;
}

int gpio_direction_output(unsigned gpio, int value)
{
	struct pinregs *regs;
	struct sharepin_cfg_addr *sc;
	int bf;

	if (gpio > GPIO_RANGE)
		return -EINVAL;
	sc = rts_get_pinaddr(gpio);
	regs = (struct pinregs *)sc->pinaddr;
	bf = gpio - sc->pinl;

	if (value)
		set_bit(bf, &regs->gpio_value);
	else
		clear_bit(bf, &regs->gpio_value);
	rts_gpio_set_field(&regs->gpio_oe, 1, 1, bf);

	return 0;
}
